import logging
import time
from lchttp.json import json_dumps
from os import scandir

from .settings import (
    portal_app,
    cache_key,
    CONSOLE_GID,
    CONSOLE_UID,
    REDIS_SOCKET,
    TMP_DIR,
)

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------
class PersistLogs:
    """
    Handle saving log lines to shared memory, and
    POSTing notifications to relevant Console API endpoint.

    Also ensure that logs are being processed in a timely
    manner. If not, we should sleep rather than process
    new loglines, to let the console catch up.
    """
    console_app = ''
    console_model = ''
    task_name = ''

    def __init__(self):
        self.queue_path = f'{TMP_DIR}/{self.console_app}/{self.console_model}'

    def __str__(self):
        return f'{self.__class__.__name__} for {self.console_app} / {self.console_model}'

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    def save(self, lines):
        """
        Takes log lines from log tailer and buffer
        them and handle post to console API.
        """
        batch_name = str(time.time())
        self.add_to_queue(batch_name, batch_data=lines)
        return self.send_task()

    def add_to_queue(self, batch_name: str, batch_data: list) -> bool:
        """
        Add batch data to batch_name key, and then save batch_name to disk.
        """
        batch = f'{self.queue_path}/{batch_name}.json'

        try:
            with open(batch, 'wb') as bq:
                bq.write(json_dumps(batch_data))
            return True

        except Exception as e:
            logger.exception(f'Error saving batch of {len(batch_data)} lines to {batch}: {e}')

        return False

    def send_task(self):
        """
        Send task scheduler the batches that are available for processing.

        Logtailer could be running, but task scheduler not running, so walk
        directory every time and send all batches found there.

        If task scheduler is running but receives a new task for batch already
        being processed, we'll leave it to the task scheduler to handle the
        duplicate task correctly, and clean up the batch file.
        """
        batches = []
        for batch_file in scandir(f'{self.queue_path}'):
            if batch_file.name.endswith('.json'):
                batches.append(batch_file.name)

        try:
            resp = portal_app.send_task(self.task_name, (batches,))
            logger.info(f'Sent batch {batches!r} to task {self.task_name!r}')
            return resp
        except Exception as e:
            logger.exception(f'Unable to send task {self.task_name!r} with args {batches!r}: {e}')
            return


# ----------------------------------------------------------------------
class PersistRedwoodLogs(PersistLogs):
    console_app = 'reporter'
    console_model = 'logline'
    task_name = 'reporter.tasks.logs.process_redwood_log_lines'


# ----------------------------------------------------------------------
class PersistDNSLogs(PersistLogs):
    console_app = 'firewall'
    console_model = 'rpzlog'
    task_name = 'reporter.tasks.logs.process_dns_log_lines'


__all__ = (
    'PersistLogs',
    'PersistRedwoodLogs',
    'PersistDNSLogs',
)
